/************************************************************************
 * Program:        GLgalaxy
 * File:           glgalaxy.c
 * Date Created:   2/01
 * Orig Author:    Uli Siegmund for Amiga
 * OpenGL Author:  Jeremy Witt (Escher) wittj@onramp.net
 *                 www.muppetclan.com - www.cryptnet1.net
 * Desc:           This is an implementation of the Classic Galaxy 
 *                 Screensaver which has been ported to a 3d Environment
 *                 using OpenGL
 *
 * Historical Information:
 *
 * Originally done by Uli Siegmund <uli@wombat.okapi.sub.org> on Amiga
 *   for EGS in Cluster
 * Port from Cluster/EGS to C/Intuition by Harald Backert
 * Port to X11 and incorporation into xlockmore by Hubert Feyrer
 *   <hubert.feyrer@rz.uni-regensburg.de>
 * Port to OpenGL and the wide world of 3D rendering by Jeremy Witt
 *   <wittj@onramp.net>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 15-Feb-01: wittj@onramp.net Ported to OpenGL + various enhancements
 * 26-Aug-00: robert.nagtegaal@phil.uu.nl and roland@tschai.demon.nl:
 *            various improvements
 * 10-May-97: jwz@jwz.org: turned into a standalone program.
 * 18-Apr-97: Memory leak fixed by Tom Schmidt <tschmidt@micron.com>
 * 07-Apr-97: Modified by Dave Mitchell <davem@magnet.com>
 * 23-Oct-94: Modified by David Bagley <bagleyd@bigfoot.com>
 *  random star sizes
 *  colors change depending on velocity
 * 10-Oct-94: Add colors by Hubert Feyer
 * 30-Sep-94: Initial port by Hubert Feyer
 * 09-Mar-94: VMS can generate a random number 0.0 which results in a
 *            division by zero, corrected by Jouk Jansen
 *            <joukj@crys.chem.uva.nl>
 *
 ***********************************************************************/
#include <X11/Intrinsic.h>

extern XtAppContext app;

#define PROGCLASS "glgalaxy"
#define HACK_INIT init_GLgalaxy
#define HACK_DRAW draw_GLgalaxy
#define HACK_RESHAPE resize_GLgalaxy
/*#define HACK_FREE release_GLgalaxy*/ /*some weirdness here.. guess this
                                         is not quite a destructor function*/
#define GLgalaxy_opts xlockmore_opts
#define DEFAULTS "*delay:         100        \n" \
                 "*showFPS:       False      \n" \
                 "*rendermode:    1          \n" \
                 "*hititerations: 400        \n" \
                 "*maxstars:      1500       \n" \
                 "*roty:          0.01       \n" \
                 "*rotx:          0.004      \n" \
                 "*ssize:         -1         \n" \
                 "*zoom:          -1         \n"

#ifdef USE_GL

#ifdef HAVE_XPM
# include <X11/xpm.h>
# ifndef PIXEL_ALREADY_TYPEDEFED
# define PIXEL_ALREADY_TYPEDEFED
# endif
#endif

#undef countof
#define countof(x) (sizeof((x))/sizeof((*x)))

#include "xlockmore.h"
 
#include <stdio.h>
#include <GL/glu.h>

#include "xpm-ximage.h"
#include "../images/star.xpm"

/*************************************************
 Defines and Macros Needed for GALAXY Algorithm
 -Modified somewhat for my use- Then renamed so 
  I could jam them back into xss
 *************************************************/
#define JLRAND()			((long) (random() & 0x7fffffff))
#define JNRAND(n)		  ((int) (JLRAND() % (n)))
#define JMAXRAND			(2147483648.0) /* unsigned 1<<31 as a float */

#define SINF(n)			  ((float)sin((double)(n)))
#define COSF(n)			  ((float)cos((double)(n)))
#define FABSF(n)		  ((float)fabs((double)(n)))

#undef MAX
#undef MIN
#undef ABS
#define MAX(a,b)((a)>(b)?(a):(b))
#define MIN(a,b)((a)<(b)?(a):(b))
#define ABS(a)((a)<0 ? -(a) : (a))

#define FLOATRAND ((double) JLRAND() / ((double) JMAXRAND))

#define MINSIZE        1
#define MINGALAXIES    2
#define MAX_IDELTAT    50
#define EPSILON 0.00000001
#define sqrt_EPSILON 0.0001
#define DELTAT (MAX_IDELTAT * 0.0001)
#define GALAXYRANGESIZE  0.1
#define GALAXYMINSIZE  0.15
#define QCONS    0.001
#define COLORBASE  256
#define RMODE_CLASSIC 0
#define RMODE_3D      1

#define RADTODEG 180/M_PI

/***********************************************************
 Structures needed by Galaxy Algorithm
 ***********************************************************/
typedef struct 
{
     double  pos[3], vel[3];
} Star;

typedef struct 
{
     int     mass;
     int     nstars;
     Star    *stars;
     double  pos[3], vel[3];
     int     galcolr;
     int     galcolb;
     int     galcolg;
} Galaxy;

typedef struct 
{
     double  mat[3][3]; /* Movement of stars(?) */
     double  scale; /* Scale */
     double  size; /* */
     double  diff[3]; /* */
     Galaxy  *galaxies; /* the Whole Universe */
     int     ngalaxies; /* # galaxies */
     int     f_hititerations; /* # iterations before restart */
     int     step;
     double  rot_y; /* rot of eye around center of universe, around y-axis*/
     double  rot_x; /* rot of eye around center of universe, around x-axis */
} unistruct;

static unistruct *universes = NULL;

/************************************************************ 
 Stuff about our window grouped together 
 ************************************************************/

typedef struct
{
     GLXContext *glx_context;
} glgalaxystruct;

static glgalaxystruct *glgalaxy = NULL;


/***********************************************************
 Globals and Command Line stuff
 ***********************************************************/

GLfloat zoom;                   /* Zoom */

GLuint  texture[1];             /* Texture pointer */
GLuint  starlist;               /* GLList for stars */
GLfloat tw;                     /* Size of stars*/

static int  _rendermode;        /* Classical (0) or 3D (1) */
static int  _hititerations;     /* number of hits before refresh */
static int  _maxstars;          /* Maximum stars in universe  */
static float _roty;             /* Universe Rotation about Y*/
static float _rotx;             /* Universe Rotation about X*/
static float _ssize;            /* Size of Stars from Command Line */
static float _zoom;             /* ZOOM from command line */
Bool         _morehelp;         /* User wants the docs instead of saver!*/


static XrmOptionDescRec opts[] = 
{
     { "-rendermode",".glgalaxy.rendermode", XrmoptionSepArg, (caddr_t) NULL },
     { "-hititerations",".glgalaxy.hititerations", 
       XrmoptionSepArg, (caddr_t) NULL},
     { "-maxstars", ".glgalaxy.maxstars", XrmoptionSepArg, (caddr_t) NULL},
     { "-roty", ".glgalaxy.roty", XrmoptionSepArg, (caddr_t) NULL},
     { "-rotx", ".glgalaxy.rotx", XrmoptionSepArg, (caddr_t) NULL},
     { "-ssize", ".glgalaxy.ssize", XrmoptionSepArg, (caddr_t) NULL},
     { "-zoom", ".glgalaxy.zoom", XrmoptionSepArg, (caddr_t) NULL},
     { "-morehelp", ".glgalaxy.morehelp", XrmoptionNoArg, "True"}
};

static argtype vars[] = 
{
     { (caddr_t *) &_rendermode, "rendermode", "Rendermode","1", t_Int},
     { (caddr_t *) &_hititerations, "hititerations", "Hititerations", "200",
       t_Int},
     { (caddr_t *) &_maxstars, "maxstars", "Maxstars", "1500", t_Int},
     { (caddr_t *) &_roty, "roty", "Roty", "0.01", t_Float},
     { (caddr_t *) &_rotx, "rotx", "Rotx", "0.004", t_Float},
     { (caddr_t *) &_ssize, "ssize", "Ssize", "-1", t_Float},
     { (caddr_t *) &_zoom, "zoom", "Zoom", "-1", t_Float},
     { (caddr_t *) &_morehelp, "morehelp", "Morehelp", "False", t_Bool}
};

static OptionStruct desc[] =
{
    {"-rendermode num", 
      "0 == classic perspective (2d) 1 == 3d perspective (1)"},
    {"-hititerations", "number of movement cycles before refresh (400)"},
    {"-maxstars", "maximum stars in universe (1500)"},
    {"-roty", "rotation factor of universe on Y axis (0.01)"},
    {"-rotx", "rotation factor of universe on X axis (0.004)"},
    {"-ssize", "Size of individual stars (-1 = Auto) range<0,1..>"},
    {"-zoom", "Zoom (-1 = Auto) range<..0,-100..>"},
    {"-morehelp","Displays help text"}
};

ModeSpecOpt GLgalaxy_opts = {countof(opts),opts, countof(vars), vars, 
                             desc};

/**********************************************************
 Function Prototypes
**********************************************************/
static void startover(void);
static void free_galaxies(unistruct *);
void        init_galaxy(void);
void        draw_galaxy(void);
Bool        LoadGLTextures(ModeInfo *);
void        displayHelp(void);

/***********************************************************
 free_galaxies
 -free memory used by galaxy structure
 ***********************************************************/
static void
free_galaxies(unistruct * gp)
{
     if (gp->galaxies != NULL) 
     {
          int         i;
          for (i = 0; i < gp->ngalaxies; i++) 
          {
               Galaxy *gt = &gp->galaxies[i];
       
               if (gt->stars != NULL)
                    (void) free((void *) gt->stars);
          }
          (void) free((void *) gp->galaxies);
          gp->galaxies = NULL;
     }
}

/***********************************************************
 startover
 -reinitializes the universe (if only it were this easy)
 ***********************************************************/
static void
startover()
{
     unistruct  *gp = universes;
     int         i, j;          /* tmp */
     double      w1, w2;        /* more tmp */
     double      d, v, w, h;    /* yet more tmp */

     gp->step = 0;
     gp->rot_y = 0;
     gp->rot_x = 0;
    
     gp->ngalaxies = JNRAND(5);  /* Random galaxies from MIN to 5 */
     if (gp->ngalaxies < -MINGALAXIES)
          gp->ngalaxies = JNRAND(-gp->ngalaxies - MINGALAXIES + 1) 
               + MINGALAXIES;
     else if (gp->ngalaxies < MINGALAXIES)
          gp->ngalaxies = MINGALAXIES;
     if (gp->galaxies == NULL)
          gp->galaxies = (Galaxy *) calloc(gp->ngalaxies, sizeof (Galaxy));
 
     for (i = 0; i < gp->ngalaxies; ++i) 
     {
          Galaxy     *gt = &gp->galaxies[i];
          double      sinw1, sinw2, cosw1, cosw2;

          gt->galcolr= JNRAND(COLORBASE - 2);
          gt->galcolb= JNRAND(COLORBASE - 2);
          gt->galcolg= JNRAND(COLORBASE - 2);

          if (gt->stars != NULL) 
          {
               (void) free((void *) gt->stars);
               gt->stars = NULL;
          }
          gt->nstars = (JNRAND(_maxstars / 2)) + _maxstars / 2;
          gt->stars = (Star *) malloc(gt->nstars * sizeof (Star));

          w1 = 2.0 * M_PI * FLOATRAND;
          w2 = 2.0 * M_PI * FLOATRAND;
          sinw1 = SINF(w1);
          sinw2 = SINF(w2);
          cosw1 = COSF(w1);
          cosw2 = COSF(w2);

          gp->mat[0][0] = cosw2;
          gp->mat[0][1] = -sinw1 * sinw2;
          gp->mat[0][2] = cosw1 * sinw2;
          gp->mat[1][0] = 0.0;
          gp->mat[1][1] = cosw1;
          gp->mat[1][2] = sinw1;
          gp->mat[2][0] = -sinw2;
          gp->mat[2][1] = -sinw1 * cosw2;
          gp->mat[2][2] = cosw1 * cosw2;

          gt->vel[0] = FLOATRAND * 2.0 - 1.0;
          gt->vel[1] = FLOATRAND * 2.0 - 1.0;
          gt->vel[2] = FLOATRAND * 2.0 - 1.0;
          gt->pos[0] = -gt->vel[0] * DELTAT * gp->f_hititerations 
               + FLOATRAND -0.5;
          gt->pos[1] = -gt->vel[1] * DELTAT * gp->f_hititerations 
               + FLOATRAND -0.5;
          gt->pos[2] = -gt->vel[2] * DELTAT * gp->f_hititerations 
               + FLOATRAND -0.5;

          gt->mass = (int) (FLOATRAND * 1000.0) + 1;
          
          gp->size = GALAXYRANGESIZE * FLOATRAND + GALAXYMINSIZE;
          
          for (j = 0; j < gt->nstars; ++j) 
          {
               Star       *st = &gt->stars[j];
               
               double      sinw, cosw;

               w = 2.0 * M_PI * FLOATRAND;
               sinw = SINF(w);
               cosw = COSF(w);
               d = FLOATRAND * gp->size;
               h = FLOATRAND * exp(-2.0 * (d / gp->size)) / 5.0 * gp->size;
               if (FLOATRAND < 0.5)
                    h = -h;
               st->pos[0] = gp->mat[0][0] * d * cosw + 
                            gp->mat[1][0] * d * sinw +
                            gp->mat[2][0] * h + gt->pos[0];

               st->pos[1] = gp->mat[0][1] * d * cosw + 
                            gp->mat[1][1] * d * sinw +
                            gp->mat[2][1] * h + gt->pos[1];

               st->pos[2] = gp->mat[0][2] * d * cosw + 
                            gp->mat[1][2] * d * sinw +
                            gp->mat[2][2] * h + gt->pos[2];
           
               v = sqrt(gt->mass * QCONS / sqrt(d * d + h * h));

               st->vel[0] = -gp->mat[0][0] * v * sinw + 
                             gp->mat[1][0] * v * cosw +
                             gt->vel[0];

               st->vel[1] = -gp->mat[0][1] * v * sinw + 
                             gp->mat[1][1] * v * cosw +
                             gt->vel[1];

               st->vel[2] = -gp->mat[0][2] * v * sinw + 
                             gp->mat[1][2] * v * cosw +
                             gt->vel[2];
           
               st->vel[0] *= DELTAT;
               st->vel[1] *= DELTAT;
               st->vel[2] *= DELTAT;
          }     
     }
}

/***********************************************************
 init_galaxy
 -called once upon program init (during initGL)
 ***********************************************************/
void
init_galaxy()
{
     unistruct  *gp;
     
     if (universes == NULL) 
     {
          if ((universes = (unistruct *) malloc(sizeof (unistruct))) == NULL)
               return;
     }
     gp = universes;

     gp->f_hititerations = _hititerations;

     switch(_rendermode)
     {
     case RMODE_CLASSIC:
          gp->scale = 12.5;
          break;
     case RMODE_3D:
          gp->scale = 37.5;
          break;
     }
     startover();
}

/***********************************************************
 draw_galaxy
 -Called for every frame, updates and draws Galaxy data
  called from drawGL
 ***********************************************************/
void
draw_galaxy()
{
     unistruct  *gp = universes;
     double      d, eps, cox=0, six=0, cor=0, sir=0;  /* tmp */
     int         i, j, k; /* more tmp */

     GLfloat     xproj,yproj;   /* used for X,Y projection of XYZ stuff */
     
     gp->rot_y += _roty;
     gp->rot_x += _rotx;

     /*only calculate projection rotation stuff if we're doing
       classical mode*/
     if (_rendermode == RMODE_CLASSIC)
     {
          cox = COSF(gp->rot_y);
          six = SINF(gp->rot_y);
          cor = COSF(gp->rot_x);
          sir = SINF(gp->rot_x);
     }

     eps = 1/(EPSILON * sqrt_EPSILON * DELTAT * DELTAT * QCONS);

     for (i = 0; i < gp->ngalaxies; ++i) 
     {
          Galaxy     *gt = &gp->galaxies[i];
          glColor4ub(gt->galcolr,gt->galcolg,gt->galcolb,255);

          for (j = 0; j < gp->galaxies[i].nstars; ++j) 
          {
               Star       *st = &gt->stars[j];
               double      v0 = st->vel[0];
               double      v1 = st->vel[1];
               double      v2 = st->vel[2];

               for (k = 0; k < gp->ngalaxies; ++k) 
               {
                    Galaxy     *gtk = &gp->galaxies[k];
                    double      d0 = gtk->pos[0] - st->pos[0];
                    double      d1 = gtk->pos[1] - st->pos[1];
                    double      d2 = gtk->pos[2] - st->pos[2];

                    d = d0 * d0 + d1 * d1 + d2 * d2;
                    if (d > EPSILON)
                         d = gt->mass / (d*sqrt(d))*DELTAT*DELTAT*QCONS;
                    else
                         d = gt->mass * eps;
                    v0 += d0 * d;
                    v1 += d1 * d;
                    v2 += d2 * d;
               }

               st->vel[0] = v0;
               st->vel[1] = v1;
               st->vel[2] = v2;
               
               st->pos[0] += v0;
               st->pos[1] += v1;
               st->pos[2] += v2;

               glLoadIdentity();
               glTranslatef(0.0f,0.0f,zoom);
               switch (_rendermode)
               {
               case RMODE_CLASSIC:
                   /*Make the projection onto a 2d surface 
                     (it just looks 3d)
                     This was converted from screen coord 
                     version to this one*/

                    xproj = (GLfloat) (((cox * st->pos[0]) - 
                                        (six * st->pos[2])) * gp->scale);
                    yproj = (GLfloat) (((cor * st->pos[1]) - 
                                        (sir * ((six * st->pos[0]) +
                                         (cox * st->pos[2])))) * gp->scale);
                    glTranslatef(xproj,yproj,0.0f); /* move to sprite loc */
                    break;
               case RMODE_3D:
                    /* rotate universe */
                    glRotatef(gp->rot_y*RADTODEG,0.0f,1.0f,0.0f);
                    glRotatef(gp->rot_x*RADTODEG,1.0f,0.0f,0.0f);
                    /* Translate to Sprite Location */
                    glTranslatef(st->pos[0],st->pos[1],st->pos[2]);
                    /* Cancel rotation of Sprite (so it faces user)*/
                    glRotatef(-gp->rot_x*RADTODEG,1.0f,0.0f,0.0f);
                    glRotatef(-gp->rot_y*RADTODEG,0.0f,1.0f,0.0f);
                    break;
               }
               glCallList(starlist); /* draw sprite */
          }
          for (k = i + 1; k < gp->ngalaxies; ++k) 
          {
               Galaxy     *gtk = &gp->galaxies[k];
               double      d0 = gtk->pos[0] - gt->pos[0];
               double      d1 = gtk->pos[1] - gt->pos[1];
               double      d2 = gtk->pos[2] - gt->pos[2];

               d = d0 * d0 + d1 * d1 + d2 * d2;
               if (d > EPSILON)
                    d = gt->mass * gt->mass / (d*sqrt(d)) * DELTAT * QCONS;
               else
                    d = gt->mass*gt->mass/(EPSILON*sqrt_EPSILON)*DELTAT*QCONS;

               d0 *= d;
               d1 *= d;
               d2 *= d;
               gt->vel[0] += d0 / gt->mass;
               gt->vel[1] += d1 / gt->mass;
               gt->vel[2] += d2 / gt->mass;
               gtk->vel[0] -= d0 / gtk->mass;
               gtk->vel[1] -= d1 / gtk->mass;
               gtk->vel[2] -= d2 / gtk->mass;
          }
          
          gt->pos[0] += gt->vel[0] * DELTAT;
          gt->pos[1] += gt->vel[1] * DELTAT;
          gt->pos[2] += gt->vel[2] * DELTAT;
  
     }

     gp->step++;
     if (gp->step > gp->f_hititerations * 4)
     {
          free_galaxies(gp);
          startover();
     }
}

/***********************************************************
 release_GLgalaxy
 -called once during program exit, cleans up memory
 ***********************************************************/
void
release_GLgalaxy(ModeInfo *mi)
{
     if (universes != NULL) 
     {
          free_galaxies(universes);
          (void) free((void *) universes);
          universes = NULL;
     }
     FreeAllGL(mi);
}


/***********************************************************
 LoadGLTextures
 -calls LoadBMP and converts to GL textures and loads them
 -...well used to anyway.. Now it calls xpm_to_ximage so XPMs are
  necessary [but a BMP floating in the DIR is not]
 ***********************************************************/
Bool LoadGLTextures(ModeInfo *mi)
{
     Bool status;
     XImage *texti2;
     
     status = False;
     texti2 = malloc(sizeof(XImage));
     if ((texti2 = xpm_to_ximage(mi->dpy,mi->xgwa.visual,
                                 mi->xgwa.colormap,starxpm)))
     {
                 status = True;
                 glGenTextures(1, &texture[0]); /* create the texture */
                 glBindTexture(GL_TEXTURE_2D, texture[0]);
                 /* actually generate the texture */
                 glTexImage2D(GL_TEXTURE_2D, 0, 4, 
                              texti2->width, texti2->height, 0,
                              GL_RGBA, GL_UNSIGNED_BYTE, texti2->data);
                 /* use linear filtering */
                 glTexParameteri(GL_TEXTURE_2D, 
                                 GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                 glTexParameteri(GL_TEXTURE_2D, 
                                 GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     }
     else
     {
          printf("Ugg! couldn't load the texture\n");
     }
     return status;
}


/***********************************************************
 resize_GLgalaxy 
 -called when window is resized
 ***********************************************************/
void resize_GLgalaxy(ModeInfo *mi, int width, int height)
{    
     /* Prevent A Divide By Zero If The Window Is Too Small */
     if (height == 0)
          height = 1;
     /* Reset The Current Viewport And Perspective Transformation */
     glViewport(0, 0, width, height);
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
     gluPerspective(45.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
     glMatrixMode(GL_MODELVIEW);
}

/***********************************************************
 init_vars
 -used to set up Zoom and StarSize based on rendering mode
 ***********************************************************/
void init_vars(int rmode)
{
     switch (rmode)
     {
     case RMODE_CLASSIC:
          if ( _zoom == -1)
               zoom = -100.0f;
          else
               zoom = (GLfloat) _zoom;
          if ( _ssize == -1)
               tw   =  0.35f;
          else
               tw   = (GLfloat) _ssize;
          break;
     case RMODE_3D:
          if ( _zoom == -1)
               zoom = -6.0f;
          else
               zoom = (GLfloat) _zoom;
          if ( _ssize == -1)
               tw   = 0.03f;
          else
               tw   = (GLfloat) _ssize;
          break;
     }
}

/***********************************************************
 init_GLgalaxy
 -GL initialization function calls init_galaxy
 ***********************************************************/
void init_GLgalaxy(ModeInfo *mi)
{
     int screen = MI_SCREEN(mi);
     glgalaxystruct *gs;
     if (!_morehelp)
     {
          if (glgalaxy == NULL)
          {
               if ((glgalaxy = (glgalaxystruct *)
                    calloc(MI_NUM_SCREENS(mi), sizeof(glgalaxystruct)))==NULL)
                    return;
          }
          gs = &glgalaxy[screen];
          
          if ((gs->glx_context = init_GL(mi)) != NULL)
          {
               if (!LoadGLTextures(mi))
               {    
                    printf("GONK! couldn't load Star texture\n");
                    return;
               }

               /*setup our Global Vars based on the mode*/
               init_vars(_rendermode);
     
               starlist=glGenLists(1);
               glNewList(starlist,GL_COMPILE);    
               glBegin(GL_TRIANGLE_STRIP);
               glTexCoord2f(1.0f, 1.0f); glVertex3f( tw, tw, 0.0f);
               glTexCoord2f(0.0f, 1.0f); glVertex3f(-tw, tw, 0.0f);
               glTexCoord2f(1.0f, 0.0f); glVertex3f( tw,-tw, 0.0f);
               glTexCoord2f(0.0f, 0.0f); glVertex3f(-tw,-tw, 0.0f);
               glEnd();
               glEndList();

               /*Setup other junk*/
               glEnable(GL_TEXTURE_2D);
               glShadeModel(GL_SMOOTH);
               glClearColor(0.0f, 0.0f, 0.0f, 0.5f);
               glClearDepth(1.0f);
          
               glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
          
               glBlendFunc(GL_SRC_ALPHA,GL_ONE);
               glEnable(GL_BLEND);

               init_galaxy();
          
               /* resizeGLScene once to set up initial perspective */
               resize_GLgalaxy(mi, MI_WIDTH(mi), MI_HEIGHT(mi));
               glFlush();
          }
          else
          {
               MI_CLEARWINDOW(mi);
          }
     }
     else
     {
          displayHelp();
          exit(0);
     }
}


/***********************************************************
 draw_GLgalaxy
 -Called once per frame, calls draw_galaxy, draws FPS if active
 ***********************************************************/
void draw_GLgalaxy(ModeInfo *mi)
{
     Window window = MI_WINDOW(mi);
     Display *dpy = MI_DISPLAY(mi);
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     
     glBindTexture(GL_TEXTURE_2D, texture[0]);
     draw_galaxy();
     if (mi->fps_p) 
          do_fps (mi);
     glXSwapBuffers(dpy, window);
     return;    
}

/***********************************************************
 displayHelp
 -Show the description texts
 ***********************************************************/
void displayHelp(void)
{
     printf("\n\nglgalaxy - Jeremy Witt (wittj@onramp.net)\n");
     printf("*****************************************\n\n");
     printf("Options: (Defaults)\n");
     printf("-rendermode      - Used to choose Rendering Capabilities\n");
     printf("                   0 == Classic Projection, (1) == 3D Mode\n");
     printf("-hititerations   - Number of cycles before refresh (400)\n");
     printf("-maxstars        - Max stars in the universe (1500)\n");
     printf("-roty            - Rotation of universe about Y axis (0.01)\n");
     printf("-rotx            - Rotation of universe about X axis (0.004)\n");
     printf("-ssize           - Used to choose size of star objects\n");
     printf("                   (-1) == Auto, interesting values [0,1]\n");
     printf("-zoom            - Used to choose zoom level\n");
     printf("                   (-1) == Auto, Valuess for Rmode 1 [-20,0]\n");
     printf("                   Valuess for Rmode 0 [-100,-15]\n");
     printf("-morehelp        - This screen\n");
     printf("-fps, (-no-fps)  - Display FPS information\n");
}

#endif /*USE_GL*/
